import re, os, subprocess, datetime, hashlib, struct

# Boot image header fields definitions.
BOOTEOM_HDR_TIMESTAMP_OFFSET = 0x4C
BOOTROM_HDR_FSBL_CHECKSUM_OFFSET = 0x50

# ELF file header fields definitions.
ELF_FHDR_PHDR_PTR_OFFSET = 0x1C
ELF_PHDR_SEG_PTR_OFFSET = 0x04
ELF_PHDR_SEG_SIZE_OFFSET = 0x10

# Regular expression to match the FSBL ELF file path.
pattern = re.compile(r'(?<=\s\[bootloader\])(\w:)'
                     r'([\\|/]\S+[\\|/])(\S+.(elf))')

# Search for FSBL ELF file path in bootgen_input.bif using regex
with open('bootgen_input.bif', 'r') as file:
    fsbl_path = re.search(pattern, file.read()).group()

with open(fsbl_path, 'rb') as file:
    # Read program header table's file offset in bytes.
    file.seek(ELF_FHDR_PHDR_PTR_OFFSET)
    e_phoff = int.from_bytes(file.read(4), byteorder='little')
    
    # Read the beginning of the data segment offset in bytes.
    file.seek(e_phoff + ELF_PHDR_SEG_PTR_OFFSET)
    p_offset = int.from_bytes(file.read(4), byteorder='little')
    
    # Read the number of bytes of the data segment.
    file.seek(e_phoff + ELF_PHDR_SEG_SIZE_OFFSET)
    p_filesz = int.from_bytes(file.read(4), byteorder='little')
    
    # Read the FSBK data from the ELF.
    file.seek(p_offset)
    fsbl_data = file.read(p_filesz)

# Calculate the md5 checksum of the FSBL data.
fsbl_checksum = hashlib.md5(fsbl_data)

# Get current timestamp.
timestamp = datetime.datetime.now()

# Convert endianness to little-endian and format into hex string.
timestamp_hex_str = format(struct.unpack('<I',
    struct.pack('>I', int(timestamp.timestamp())))[0], '08x')

# Write the FSBL checksum, timestamp, and Image Validity Word
# to UDF file.
with open('udf.txt', 'w') as file:
    file.write(timestamp_hex_str)
    file.write('{:032x}'.format(int.from_bytes(
        fsbl_checksum.digest(), byteorder='big')))
    file.write('ffffffff')

# Specify a file name of the boot image.
filename = 'BOOT4.bin'

# Specify a file path to the bootgen.bat
bootgen_path = r'D:\Xilinx\SDK\2018.2\bin\bootgen.bat'

# Specify arguments to be passed to the bootgen.bat
# -image <filename>[.bif] - specifies the input BIF file.
# -arch [zynq | zynqmp]   - Xilinx architecture
# -o <filename>           - specifies  the output BIN file.
#                           Use no or .bin extension.
# -w [on | off]           - overwrite mode
bootgen_args = r'-image {cwd}\bootgen_input.bif ' \
               r'-arch zynq ' \
               r'-o {cwd}\{filename} ' \
               r'-w on'.format(cwd=os.getcwd(), filename=filename)

# Invoke the bootgen.bat to build a boot image.
subprocess.run([bootgen_path, bootgen_args])